import werkzeug, os, sys
import subprocess
import json
import uuid
import time
import urllib.parse
import http.client
from jinja2 import Environment, FileSystemLoader
from flask import Flask, send_file, make_response, request, copy_current_request_context
from threading import Thread
from flask_restplus import Resource, Api, fields, reqparse
from stat import *
from flask_cors import CORS
from jsonschema import validate
from jsonschema import Draft7Validator

from pymongo import MongoClient
import bson.objectid
from bson.json_util import dumps


app = Flask(__name__)
CORS(app)

app.config.SWAGGER_UI_JSONEDITOR = True
app.config.SWAGGER_UI_DOC_EXPANSION = 'list'

api = Api(app, doc='/doc/')

REGISTRY_CERTS_UPLOAD_FOLDER = '/home/ubuntu/terraform/scripts/certs/registry'

parser = reqparse.RequestParser()
parser.add_argument('file', type=werkzeug.datastructures.FileStorage, location='files', required=True)


env_fields = api.model('EnvModel', {
    'cloudbroker_endpoint': fields.String(description='The Cloudbroker REST Api endpoint', required=True),
    'cloudbroker_username': fields.String(description='The username', required=True),
    'cloudbroker_password': fields.String(description='The password', required=True),
    'repository_user': fields.String(description='The Git repo username', required=True),
    'repository_pwd': fields.String(description='The Git repo password', required=True),
    'repository_url': fields.String(description='The Git repo http(s) url', required=True),
    'revision': fields.String(description='The code revision for the configuration', required=False),
    'filepath': fields.String(description='The filepath for the configuration file in the repo', required=True),
    'monitoring_host': fields.String(description='The ADAPT Monitoring REST Api endpoint url', required=True),
    'monitoring_port': fields.String(description='The ADAPT Monitoring REST Api endpoint port', required=True),
})

schemaFile = open('schemas/application_description.schema.json', 'r')
schemaContent = schemaFile.read()
appDescriptionSchema = json.loads(schemaContent)

@app.route("/")
def main():
    index_path = os.path.join(app.static_folder, 'index.html')
    return send_file(index_path)

@api.route("/schema")
class DumpSchema(Resource):
    def get(self):
        print(json.dumps(api.__schema__, indent=4, sort_keys=False))
        return json.dumps(api.__schema__)
        
@api.route("/certs/registry/<registryIp>")
@api.doc(params={'registryIp': 'The ip of the docker registry for which we want to upload a certificate'})
class RegistryCertsUpload(Resource):
    decorators=[]
    def post(self, registryIp):
        print("Got request to create certs for ip " + registryIp)
        respData = {}

        data = parser.parse_args()
        if data['file'] == "":
            respData["content"] = "No file found"
            respData["resultCode"] = "400"
            respData["result"] = "error"
            response = make_response(json.dumps(respData))
            response.status_code = 400
            return response

        cert = data['file']

        if cert:
            filename = 'ca.crt'

            if uploadRegistryCerts(cert, registryIp, filename):
                respData["resultCode"] = "200"
                respData["result"] = "success"
                response = make_response(json.dumps(respData))
                response.status_code = 200
                return response

        respData["content"] = "Something went wrong"
        respData["resultCode"] = "400"
        respData["result"] = "error"
        response = make_response(json.dumps(respData))
        response.status_code = 400
        return response

        
def uploadRegistryCerts(cert, registryIp, filename):
    try:
        destinationFolder = os.path.join(REGISTRY_CERTS_UPLOAD_FOLDER,registryIp)
        if not os.path.exists(destinationFolder):
            os.makedirs(destinationFolder)
        print("Saving cert to " + destinationFolder)
        cert.save(os.path.join(destinationFolder,filename))
        return True
    except:
        print("Unexpected error:", sys.exc_info()[0])
        return False
        
        
@api.route("/terraform/<environment>")
@api.doc(params={'environment': 'The environment [infrastructure | services] for which we want to create the configuration files'})
class Configuration(Resource):
    @api.expect(env_fields)
    def post(self, environment):
        data = {}
        body = request.get_json()
        print(body)
        repo_user = body["repository_user"]
        repo_pwd = body["repository_pwd"]
        repo_url = body["repository_url"]
        repo_url = repo_url[8:]
        repo_url = "https://"+urllib.parse.quote(repo_user)+":"+urllib.parse.quote(repo_pwd)+"@"+repo_url
        if not "repository_url" in body or not "filepath" in body or not body["repository_user"] or not body["repository_pwd"]:
            data["result"] = "error"
            data["resultCode"] = "400"
            data["content"] = "Missing parameters"
            response = make_response(json.dumps(data))
            response.status_code = 400
            return response
        repository = body["repository_url"].split('/')[-1]
        repository = repository[:repository.index('.git')]
        filePath = body["filepath"]
        if not os.path.exists("repo/"+repository):
            print("Cloning repository: " + body["repository_url"])
            if "branch" in body:
                branch = body["branch"]
                result = subprocess.check_output(['git', 'clone', repo_url, '-b', branch], cwd="repo/")
            else:
                result = subprocess.check_output(['git', 'clone', repo_url], cwd="repo/")
        print("Pulling repository: " + body["repository_url"])    
        result = subprocess.check_output(['git', 'pull'], cwd="repo/"+repository)
        if "revision" in body:
            revision = body["revision"]
        else:
            revision = "HEAD"
        print("Getting revision " + revision + " for file " + filePath + " from repository " + repository + " (" + body["repository_url"] + ")")
        result = subprocess.check_output(['git', 'checkout', '-m', revision, filePath], cwd="repo/"+repository)


        file = open('repo/'+repository+"/"+filePath, 'r')
        content = file.read()

        print("Creating config for " + environment + " with input file:")
        print(json.loads(content))
        jsonContent = json.loads(content)
        try:
            jsonContent = json.loads(content)
            print(">>>>>>>>>>>>>>>>Validating content against schema")
            #validate(jsonContent, appDescriptionSchema)
            Draft7Validator(appDescriptionSchema).is_valid(jsonContent)
        except:
            print("*************Validation error!****************", sys.exc_info()[0], sys.exc_info()[1])
        if "cloudbroker_username" in body and "cloudbroker_password" in body and "cloudbroker_endpoint" in body and body["cloudbroker_username"] and body["cloudbroker_password"] and body["cloudbroker_endpoint"]:
            jsonContent["cloudbrokerEndpoint"] = body["cloudbroker_endpoint"]
            jsonContent["cloudbrokerUsername"] = body["cloudbroker_username"]
            jsonContent["cloudbrokerPassword"] = body["cloudbroker_password"]
            print("Changed credentials!")
        if "privateClouds" in body and body["privateClouds"]:
            jsonContent["privateClouds"] = body["privateClouds"]

        data["result"] = "success"
        data["resultCode"] = "202"

        if environment == "infrastructure":
            create_infrastructure_tf(jsonContent)
            data["content"] = "Request for infrastructure configuration sent"
            return make_response(json.dumps(data))
        elif environment == "services":
            create_services_tf(jsonContent)
            data["content"] = "Request for services configuration sent"
            return make_response(json.dumps(data))
        elif environment == "all":
            create_infrastructure_tf(jsonContent)
            create_services_tf(jsonContent)
            data["content"] = "Request for infrastructure and services configuration sent"
            return make_response(json.dumps(data))
        else:
            data["result"] = "error"
            data["resultCode"] = "400"
            data["content"] = "Wrong environment parameter. Accepted values are: 'infrastructure'|'services'|'all'"
            response = make_response(json.dumps(data))
            response.status_code = 400
            return response



@api.route("/terraform/<operation>/<appName>/<folder>")
@api.doc(params={'operation': 'The Terraform operation to perform: [init | plan | apply]', 'appName':'The name of the application','folder':'The logical application area we want to manage: [infrastructure | services]'})
class terraformOperation(Resource):
    @api.expect(env_fields)
    def post(self, operation, appName, folder):
        op_id = str(uuid.uuid4())
        body = request.get_json()
        status_data = {}
        @copy_current_request_context
        def terraformThread(operation, appName, folder, op_id, status_data):
            print("Starting thread for operation " + op_id)
            status_data["operation"] = operation
            status_data["folder"] = folder
            status_data["appName"] = appName
            status_data["operation_id"] = op_id
            log_dir = appName+"/logs"
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)
            status_data["result"] = "running"
            try:
                client = MongoClient()
                adaptDoDb = client['adapt-do-db']
                operations = adaptDoDb.operations
                jdata = json.loads(json.dumps(status_data))
                db_result = operations.update_one({'operation_id': op_id}, {'$set': jdata}, upsert=True)
                print("POST operation: %s record(s) matched, %s record(s) modified, upserted id: %s" % (
                    db_result.matched_count, db_result.modified_count, db_result.upserted_id))
            except:
                print("Unexpected error:", sys.exc_info()[0])
            finally:
                client.close()

            #result = make_response(subprocess.run(['terraform', operation, '-no-color'], cwd=appName+"/"+folder, stdout=subprocess.PIPE, stderr=subprocess.STDOUT).stdout.decode('utf-8'))
            with open(log_dir+"/"+folder+"-"+operation+"-"+op_id+".log","w") as out:
                if operation == "destroy":
                    myproc = subprocess.run(['terraform', operation, '-no-color', '-force'], cwd=appName+"/"+folder, stdout=out, stderr=out)
                else:
                    print("Starting subprocess")
                    myproc = subprocess.run(['terraform', operation, '-no-color'], cwd=appName+"/"+folder, stdout=out, stderr=out)
            print("myproc return code: " + str(myproc.returncode))
            
            #section to manage the update of the app description file upon completion of 'apply' operations
            if myproc.returncode == 0:
                status_data["result"] = "success"
            elif myproc.returncode > 0:
                status_data["result"] = "failed"

            try:
                client = MongoClient()
                adaptDoDb = client['adapt-do-db']
                operations = adaptDoDb.operations
                jdata = json.loads(json.dumps(status_data))
                db_result = operations.update_one({'operation_id': op_id}, {'$set': jdata}, upsert=True)
                print("POST operation: %s record(s) matched, %s record(s) modified, upserted id: %s" % (
                db_result.matched_count, db_result.modified_count, db_result.upserted_id))
            except:
                print("Unexpected error:", sys.exc_info()[0])
            finally:
                client.close()

            if operation == "apply":
                data = {}
                if body:
                    print("Body is present in request ")
                else: 
                    print("Body is not present in request ")
                
                if body:
                    repo_user = body["repository_user"]
                    repo_pwd = body["repository_pwd"]
                    repo_url = body["repository_url"]
                    repo_url = repo_url[8:]
                    repo_url = "https://"+urllib.parse.quote(repo_user)+":"+urllib.parse.quote(repo_pwd)+"@"+repo_url
                    print("repo_url for clone: " + repo_url)
                    if not "repository_url" in body or not "filepath" in body or not body["repository_user"] or not body["repository_pwd"]:
                        data["result"] = "error"
                        data["resultCode"] = "400"
                        data["content"] = "Missing parameters"
                        response = make_response(json.dumps(data))
                        response.status_code = 400
                        return response
                    repository = body["repository_url"].split('/')[-1]
                    repository = repository[:repository.index('.git')]
                    filePath = body["filepath"]
                    if not os.path.exists("repo/"+repository):
                        print("Cloning repository: " + body["repository_url"])
                        if "branch" in body:
                            branch = body["branch"]
                            result = subprocess.check_output(['git', 'clone', repo_url, '-b', branch], cwd="repo/")
                        else:
                            result = subprocess.check_output(['git', 'clone', repo_url], cwd="repo/")
                    print("Pulling repository: " + body["repository_url"])    
                    result = subprocess.check_output(['git', 'pull'], cwd="repo/"+repository)
                    if "revision" in body:
                        revision = body["revision"]
                    else:
                        revision = "HEAD"
                    print("Getting revision " + revision + " for file " + filePath + " from repository " + repository + " (" + body["repository_url"] + ")")
                    result = subprocess.check_output(['git', 'checkout', '-m', revision, filePath], cwd="repo/"+repository)

                    with open('repo/'+repository+"/"+filePath, 'r') as ad:
                        content = ad.read()
                        print("Got application descriptor:")
                        print(json.loads(content))
                        jsonContent = json.loads(content)
                    
                    if folder == "infrastructure":
                        #Generating a unique instance id for the application, to be saved in the app descriptor
                        print("Generating unique instance id for application")
                        jsonContent['applicationInstanceId'] = str(uuid.uuid4())
                        print("Generated unique instance id for application: " + jsonContent['applicationInstanceId'])

                        status_urls = []
                        for p in jsonContent['virtualMachines']:
                            nodeName = p['dockerHostNodeName']
                            nodeIpLabel = nodeName+".external_ip_address"
                            vm_ip = subprocess.check_output(['terraform', 'output', nodeIpLabel], cwd=appName+"/"+folder).decode(encoding='UTF-8').strip()
                            if vm_ip:
                                print("Got ip: " + vm_ip + " for node " + nodeName )
                                p['dockerHostPublicIp'] = vm_ip
                                status_urls.append(request.url_root+"vm/"+vm_ip+"/status")


                        try:
                            client = MongoClient()
                            adaptDoDb = client['adapt-do-db']
                            operations = adaptDoDb.operations
                            json_op = operations.find_one({'operation_id': op_id})
                            status_data = json.loads(json_op)
                            status_data["status_urls"] = status_urls
                            jdata = json.loads(json.dumps(status_data))
                            db_result = operations.update_one({'operation_id': op_id}, {'$set': jdata}, upsert=True)
                            print("POST operation: %s record(s) matched, %s record(s) modified, upserted id: %s" % (
                                db_result.matched_count, db_result.modified_count, db_result.upserted_id))
                        except:
                            print("Unexpected error:", sys.exc_info()[0])
                        finally:
                            client.close()

                        with open('repo/'+repository+"/"+filePath, 'w') as ad:
                            json.dump(jsonContent, ad, indent=4, sort_keys=False)
                            print("Updated locally application description:")
                            print(json.dumps(jsonContent, indent=4, sort_keys=False))
                        result = subprocess.check_output(['git', 'config', 'user.email', 'fakeemail@fakeemail.com'], cwd="repo/"+repository)
                        result = subprocess.check_output(['git', 'config', 'user.name', repo_user], cwd="repo/"+repository)
                        print("Committing application description file:")
                        try:
                            result = subprocess.check_output(['git', 'commit', '-a', '-m', 'Added node ip'], cwd="repo/"+repository)
                        except subprocess.CalledProcessError as commitError:
                            output = commitError.output.decode(encoding='UTF-8')
                            if "Your branch is up-to-date" not in output:
                                print("Commit error: " + output)
                                raise Exception("Error in commit", output)
                        print("Pushing application description file:")
                        result = subprocess.check_output(['git', 'push', repo_url], cwd="repo/"+repository)
                        print("Push done")
                        
                    if folder == "services":                        
                        ##Invoke the Monitoring API to start monitoring
                        startMonitoring(body)
                            
            if operation == "destroy" and folder == "services":
                ##Invoke the Monitoring API to stop monitoring
                if body: 
                    repository = body["repository_url"].split('/')[-1]
                    repository = repository[:repository.index('.git')]
                    filePath = body["filepath"]
                    if "revision" in body:
                        revision = body["revision"]
                    else:
                        revision = "HEAD"
                    print("Stopping monitoring for app " + body["repository_url"])
                    applicationInstanceId = ""
                    with open('repo/'+repository+"/"+filePath, 'r') as ad:
                        content = ad.read()
                        print("Got application descriptor:")
                        print(json.loads(content))
                        jsonContent = json.loads(content)
                        applicationInstanceId = jsonContent['applicationInstanceId']
                    if applicationInstanceId:
                        stopMonitoring(body, applicationInstanceId)
                    else:
                        print("Cannot stop monitoring: no applicationInstanceId found!")
                else:
                    print("Cannot stop monitoring: no input data provided")
                        
        t = Thread(target=terraformThread, args=(operation, appName, folder, op_id, status_data))
        t.start()
        data = {}
        data["result"] = 'success'
        data["operation"] = operation
        data["folder"] = folder
        data["appName"] = appName
        data["operation_id"] = op_id
        data["get_log_url"] = request.url_root + "terraform/"+operation+"/"+appName+"/"+folder+"/"+op_id
        data["get_result_url"] = request.url_root + "terraform/operation/result/"+op_id
        response = make_response(json.dumps(data))
        response.headers['Content-Type'] = 'application/json'
        response.status_code = 202
        return response


@api.route("/terraform/<operation>/<appName>/<folder>/<op_id>")
@api.doc(params={'op_id': 'The Operation ID returned by the POSTed operation request'})
class OperationLog(Resource):
    @api.doc(responses={200: 'Ok'})
    def get(self, operation, appName, folder, op_id):
        print("Reading from file: " + appName+"/logs/"+folder+"-"+operation+"-"+op_id+".log")
        with open(appName+"/logs/"+folder+"-"+operation+"-"+op_id+".log", 'r') as f:
            logdata = f.read()
            if len(logdata) > 0:
                result = "0"
            else:
                result = "-1"

        #result = subprocess.run(['tail', appName+"/logs/"+folder+"-"+operation+"-"+op_id+".log"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        data = {}
        if result == "0":
            data["result"] = "success"
        else:
            data["result"] = "failure"
        data["resultCode"] = result
        print("Output: " + str(logdata))
        data["content"] = logdata
        response = make_response(json.dumps(data))
        response.headers['Content-Type'] = 'application/json'
        response.status_code = 200
        return response

@api.route("/terraform/operation/result/old/<op_id>")
@api.doc(params={'op_id': 'The Operation ID returned by the POSTed operation request'})
class OperationStatus(Resource):
    def get(self, op_id):
        print("Reading from file: /app/operations/"+op_id+".txt")
        with open("/app/operations/"+op_id+".txt", 'r') as f:
            lines = f.read().splitlines()
            if len(lines) > 0:
                result = lines[0]
            else:
                result = "-1"

        data = {}
        print("result: " + result)
        if result == "0":
            data["result"] = "success"
        elif result == "running":
            data["result"] = "running"
        else:
            data["result"] = "failure"
        data["resultCode"] = result
        response = make_response(json.dumps(data))
        response.headers['Content-Type'] = 'application/json'
        response.status_code = 200
        return response


@api.route("/terraform/adapt-deployer/prepare")
class adaptDeploymentPreparation(Resource):
    def post(self):
        jsonContent = request.get_json()
        create_adapt_tf(jsonContent)
        return make_response("Request for 'adapt' configuration sent.\n")


@api.route("/terraform/adapt-deployer/<appName>/privateSshKey")
class adaptAddPrivateSshKey(Resource):
    def post(self,appName):
        file = request.files['file']
        if file:
            if not os.path.exists('/home/ubuntu/terraform/keypairs/'+appName):
                os.makedirs('/home/ubuntu/terraform/keypairs/'+appName)

            filename = 'private-key-openssh'
            pathname = os.path.join('/home/ubuntu/terraform/keypairs/'+appName, filename)
            file.save(pathname)
            os.chmod(pathname, S_IRUSR)
        return make_response("Request for ssh key upload sent.\n")

@api.route("/terraform/adapt-deployer/<operation>/<appName>/<folder>")
@api.doc(params={'operation': 'The Terraform operation to perform: [init | plan | apply]', 'appName':'The name of the application','folder':'The logical application area we want to manage: [infrastructure | services]'})
class terraformAdaptNodeOperation(Resource):
    def post(self, operation, appName, folder):
        op_id = str(uuid.uuid4())

        @copy_current_request_context
        def terraformThread(operation, appName, folder, op_id):
            print("Starting thread for operation " + op_id)
            log_dir = appName+"/logs"
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)
            ops_dir = "/app/operations"
            if not os.path.exists(ops_dir):
                os.makedirs(ops_dir)
            with open(ops_dir+"/"+op_id+".txt", 'w') as f:
                f.write("running")

            #result = make_response(subprocess.run(['terraform', operation, '-no-color'], cwd=appName+"/"+folder, stdout=subprocess.PIPE, stderr=subprocess.STDOUT).stdout.decode('utf-8'))
            with open(log_dir+"/"+folder+"-"+operation+"-"+op_id+".log","w") as out:
                if operation == "destroy":
                    myproc = subprocess.run(['terraform', operation, '-no-color', '-force'], cwd=appName+"/"+folder, stdout=out, stderr=out)
                else:
                    print("Starting subprocess")
                    myproc = subprocess.run(['terraform', operation, '-no-color'], cwd=appName+"/"+folder, stdout=out, stderr=out)
            print("myproc return code: " + str(myproc.returncode))
            
            #section to manage the update of the app description file upon completion of 'apply' operations
            with open(ops_dir+"/"+op_id+".txt", 'w') as f:
                f.write(str(myproc.returncode))
                        
        t = Thread(target=terraformThread, args=(operation, appName, folder, op_id))
        t.start()
        data = {}
        data["operation_id"] = op_id
        data["get_log_url"] = request.url_root + "terraform/"+operation+"/"+appName+"/"+folder+"/"+op_id
        data["get_result_url"] = request.url_root + "terraform/operation/result/"+op_id
        response = make_response(json.dumps(data))
        response.headers['Content-Type'] = 'application/json'
        response.status_code = 202
        return response

        
@api.route("/deploy/<appName>")
class deployApp(Resource):
    decorators=[]
    def post(self, appName):
        respData = {}
        uploaded_files = request.files.getlist("file")
        print("uploaded_files")
        for file in uploaded_files:
            print ("name: " + file.name)
            print ("filename: " + file.filename)
            if file.filename == "body":
                body = json.loads(file.stream.read())
                repository_user = body["repository_user"]
                repository_pwd = body["repository_pwd"]
                repository_url = body["repository_url"]
                revision = body["revision"]
                dockerRegistryIp = body["dockerRegistryIp"]
            elif file.filename == "fileCaCert":
                fileCaCert = file
            elif file.filename == "fileSshKey":
                fileSshKey = file
        
        print("Uploading CA cert for docker registry at " + dockerRegistryIp)
        if uploadRegistryCerts(fileCaCert, dockerRegistryIp, "ca.crt"):
            print("CA cert uploadaed to " + dockerRegistryIp)

        print("Preparing configuration environment for Terraform...")
        
        respData["content"] = "fileCaCert and fileSshKey found"
        respData["resultCode"] = "200"
        respData["result"] = "success"
        response = make_response(json.dumps(respData))
        response.status_code = 200
        print(response.status)
        return response



        

@app.route("/sleep/<seconds>")
def test(seconds):
    subprocess.Popen(['whoami'])
    subprocess.run(['sleep', seconds])
    result = make_response("I went to sleep...")
    result.headers['Content-Type'] = 'text/plain'
    return result

@api.route("/vm/vmlist/<vmIp>/<vmStatus>")
class adaptVmIpList(Resource):
    def post(self, vmIp, vmStatus):
        respData = {}
        vmList = {}
        print("Received post message for VM " + vmIp + ": status is " + vmStatus)
        try:
            client = MongoClient()
            adaptDoDb = client['adapt-do-db']
            vmListTable = adaptDoDb.vmList
            vmListMongo = vmListTable.find_one()
            if not vmListMongo:
                vmListMongo = {}
            else:
                del vmListMongo['_id']
            vmList = json.loads(bson.json_util.dumps(vmListMongo))
            vmList[vmIp] = vmStatus
            jdata = json.loads(bson.json_util.dumps(vmList))
            print(jdata)
            db_result = vmListTable.replace_one({}, jdata, upsert=True)
            print("POST operation: %s record(s) matched, %s record(s) modified, upserted id: %s" % (
                db_result.matched_count, db_result.modified_count, db_result.upserted_id))
            response = make_response("Vm status saved\n")
            response.status_code = 200
        except:
            print("Unexpected error:", sys.exc_info()[0], sys.exc_info()[1])
            respData["resultCode"] = "404"
            respData["result"] = "Unexpected error: %s %s" % (sys.exc_info()[0], sys.exc_info()[1])
            response = make_response(json.dumps(respData))
            response.status_code = 404
        finally:
            client.close()

        return response



@api.route("/vm/vmlist")
class adaptVmIpList(Resource):
    def get(self):
        respData = {}
        print("Received get request for VM ip list")
        try:
            client = MongoClient()
            adaptDoDb = client['adapt-do-db']
            vmListTable = adaptDoDb.vmList
            vmListMongo = vmListTable.find_one()
            if vmListMongo:
                del vmListMongo['_id']
            response = make_response(bson.json_util.dumps(vmListMongo))
            response.headers['Content-Type'] = 'application/json'
            response.status_code = 200
        except:
            print("Unexpected error:", sys.exc_info()[0])
            respData["resultCode"] = "404"
            respData["result"] = "Unexpected error: %s %s" % (sys.exc_info()[0], sys.exc_info()[1])
            response = make_response(json.dumps(respData))
            response.status_code = 404
        finally:
            client.close()

        return response


@api.route("/vm/<vmIp>/status")
class adaptVmStatus(Resource):
    def post(self, vmIp):
        respData = {}
        jsonContent = request.get_json()
        print("Received post message for VM status: " + json.dumps(jsonContent))
        print("Saving vm status:")
        try:
            client = MongoClient()
            adaptDoDb = client['adapt-do-db']
            vmStatusTable = adaptDoDb.vmStatus
            db_result = vmStatusTable.update_one({"vmPublicIp": vmIp}, {'$set': jsonContent}, upsert=True)
            responseStr = "POST operation: %s record(s) matched, %s record(s) modified, upserted id: %s" % (
                db_result.matched_count, db_result.modified_count, db_result.upserted_id)
            print(responseStr)
            response = make_response("vm status saved: " + json.dumps(jsonContent))
            response.status_code = 200
        except:
            print("Unexpected error:", sys.exc_info()[0], sys.exc_info()[1])
            respData["resultCode"] = "400"
            respData["result"] = "Unexpected error: %s %s" % (sys.exc_info()[0], sys.exc_info()[1])
            response = make_response(json.dumps(respData))
            response.status_code = 400
        finally:
            client.close()

        return response

@api.route("/vm/<vmIp>/status")
class adaptVmStatus(Resource):
    def get(self, vmIp):
        respData = {}
        print("Received get request for status of vm with IP: " + vmIp)
        try:
            client = MongoClient()
            adaptDoDb = client['adapt-do-db']
            vmStatusTable = adaptDoDb.vmStatus
            content = vmStatusTable.find_one({"vmPublicIp": vmIp})
            response = make_response(bson.json_util.dumps(content))
            response.headers['Content-Type'] = 'application/json'
            response.status_code = 200
        except:
            print("Unexpected error:", sys.exc_info()[0], sys.exc_info()[1])
            respData["resultCode"] = "400"
            respData["result"] = "Unexpected error: %s %s" % (sys.exc_info()[0], sys.exc_info()[1])
            response = make_response(json.dumps(respData))
            response.status_code = 400
        finally:
            client.close()

        return response


@api.route("/monitoringmanager/api/applications")
class adaptMonitoringSimulation(Resource):
    def post(self):
        jsonContent = request.get_json()
        print("Received START monitoring request: " + json.dumps(jsonContent))
        return make_response("Request for 'adapt monitoring' configuration received.\n")

@api.route("/monitoringmanager/api/applications/<applicationInstanceId>")
class adaptMonitoringSimulation(Resource):
    def delete(self, applicationInstanceId):
        jsonContent = request.get_json()
        print("Received STOP monitoring request for applicationInstanceId: " + applicationInstanceId)
        return make_response("Request for 'adapt monitoring' stop received.\n")

@api.route("/terraform/operation/result/<op_id>")
@api.doc(params={'op_id': 'The Operation ID returned by the POSTed operation request'})
class OperationStatus(Resource):
    def get(self, op_id):
        print("Reading from file: /app/operations/"+op_id+".json")
        data = {}
        try:
            try:
                client = MongoClient()
                adaptDoDb = client['adapt-do-db']
                operations = adaptDoDb.operations
                content = operations.find_one({'operation_id': op_id})
            except:
                print("Unexpected error:", sys.exc_info()[0], sys.exc_info()[1])
            finally:
                client.close()

            print("Got op_id json:")
            print(content)

            response = make_response(bson.json_util.dumps(content))
            response.headers['Content-Type'] = 'application/json'
            response.status_code = 200
            return response
        except:
            data["result"] = "failure"
            data["message"] = "Operation status not found"
            response = make_response(json.dumps(data))
            response.status_code = 404
            print("Unexpected error:", sys.exc_info()[0], sys.exc_info()[1])
            return response

@api.route("/test")
class MyTest(Resource):
    def get(self):
        try:
            with open("/app/myfile.json", 'r') as f:
                content = f.read()
                print("**************** content: " + content)
            with open("/app/myfile.json", 'w') as f:
                status_data = json.loads(content)
                status_urls = []
                status_urls.append("Item1")
                status_urls.append("Item2")
                status_data["status_urls"] = status_urls
                print("**************dumps: " + json.dumps(status_data))
                json.dump(status_data, f, indent=4, sort_keys=False)
        except:
            print("Unexpected error:", sys.exc_info()[0])
        #status_data = json.loads('{"aa": "bb" }')
        #status_urls = []
        #status_urls.append("Item1")
        #status_urls.append("Item2")
        #status_data["status_urls"] = status_urls
        #print("*************** status_data ***********: " + json.dumps(status_data))


@api.route("/mongotest/<id>")
class MongoTest(Resource):
    def get(self, id):
        try:
            client = MongoClient()
            adaptDoDb = client['adapt-do-db']
            operations = adaptDoDb.operations
            print("GETting operation with id: " + id)
            result = operations.find_one({'operation_id': id })
            print(result)
        except:
            print("Unexpected error:", sys.exc_info()[0])
        finally:
            client.close()
        return make_response("GET request for 'mongotest' received.\n")

@api.route("/mongotest/<id>")
class MongoTest(Resource):
    def post(self, id):
        try:
            client = MongoClient()
            adaptDoDb = client['adapt-do-db']
            operations = adaptDoDb.operations
            #op = { 'id': id, 'status': 'creating'}
            #result = operations.update_one({ 'id': id}, {'$set': op}, upsert=True)
            #print('POST operation: {0}'.format(result.inserted_id))
            operation = "init"
            folder = "infrastructure"
            appName = "ExampleApp"
            op_id = "abcfders"
            status_data = {}
            status_data["operation"] = operation
            status_data["folder"] = folder
            status_data["appName"] = appName
            status_data["operation_id"] = id
            status_data["result"] = "starting"
            operations = adaptDoDb.operations
            jdata = json.loads(json.dumps(status_data))

            #op = { 'id': id, 'status': 'snorkeling2'}
            #db_result = operations.update_one({ 'id': id}, {'$set': op}, upsert=True)

            db_result = operations.update_one({'operation_id': id}, {'$set': jdata}, upsert=True)
            print("POST operation: %s record(s) matched, %s record(s) modified, upserted id: %s" % (db_result.matched_count, db_result.modified_count, db_result.upserted_id))

        except:
            print("Unexpected error:", sys.exc_info()[0], sys.exc_info()[1])
        finally:
            client.close()
        return make_response("POST request for 'mongotest' received.\n")


# Everything not declared before (not a Flask route / API endpoint)...
@app.route('/<path:path>')
def route_frontend(path):
    # ...could be a static file needed by the front end that
    # doesn't use the `static` path (like in `<script src="bundle.js">`)
    file_path = os.path.join(app.static_folder, path)
    if os.path.isfile(file_path):
        return send_file(file_path)
    # ...or should be handled by the SPA's "router" in front end
    else:
        index_path = os.path.join(app.static_folder, 'index.html')
        return send_file(index_path)


########################
PATH = os.path.dirname(os.path.abspath(__file__))
TEMPLATE_ENVIRONMENT = Environment(
    autoescape=False,
    loader=FileSystemLoader(os.path.join(PATH, 'templates')),
    trim_blocks=True)


def render_template(template_filename, context):
    return TEMPLATE_ENVIRONMENT.get_template(template_filename).render(context)


def create_infrastructure_tf(data_loaded):

    appName = data_loaded["name"]
    consulJoinIp = data_loaded["consulJoinIp"]
    infrastructure_dir = appName+"/infrastructure"
    service_dir = appName+"/services"
    if not os.path.exists(appName):
        os.makedirs(appName)
    if not os.path.exists(infrastructure_dir):
        os.makedirs(infrastructure_dir)
    if not os.path.exists(service_dir):
        os.makedirs(service_dir)

    privateClouds = {}
    if 'privateClouds' in data_loaded:
        for privateCloud in data_loaded["privateClouds"]:
            privateClouds[privateCloud["id"]] = privateCloud
        
    for element in data_loaded["virtualMachines"]:
        print(element)
        fname = infrastructure_dir+"/"+appName+"-vm-"+element["dockerHostNodeName"]+".tf"
        if 'onPrivateCloud' in element:
            privateCloud = privateClouds[element["privateCloudId"]]
            context = {
                    'privateCloudUserName': privateCloud["user_name"],
                    'privateCloudTenantName': privateCloud["tenant_name"],
                    'privateCloudPassword': privateCloud["password"],
                    'privateCloudAuthUrl': privateCloud["auth_url"],
                    'privateCloudRegion': privateCloud["region"],
                    'appName': appName,
                    'consulJoinIp': consulJoinIp,
                    'adaptHost': request.url_root,
                    #'docker_host_ip': element["docker-host-ip"],
                    'dockerHostNodeName': element["dockerHostNodeName"],
                    'vmUser': element["vmUser"],
                    'keyPair': element["keyPair"],
                    'imageId': element["imageId"],
                    'flavorId': element["flavorId"],
                    'securityGroups': element["securityGroups"],
                    'networkName': element["networkName"],
                    'floatingIpPoolName': element["floatingIpPoolName"]
            }
            if 'dockerPrivateRegistryIp' in element:
                   context['dockerPrivateRegistryIp'] = element["dockerPrivateRegistryIp"]
            if 'dockerPrivateRegistryPort' in element:
                    context['dockerPrivateRegistryPort'] =  element["dockerPrivateRegistryPort"]

            with open(fname, 'w') as f:
                    tf = render_template('vm-openstack-tpl.tf', context)
                    f.write(tf)            
        else:
            context = {
                    'appName': appName,
                    'consulJoinIp': consulJoinIp,
                    'adaptHost': request.url_root,
                    #'docker_host_ip': element["docker-host-ip"],
                    'dockerHostNodeName': element["dockerHostNodeName"],
                    'cloudbrokerEndpoint': data_loaded["cloudbrokerEndpoint"],
                    'cloudbrokerUsername': data_loaded["cloudbrokerUsername"],
                    'cloudbrokerPassword': data_loaded["cloudbrokerPassword"],
                    'vmSoftwareId': element["vmSoftwareId"],
                    'vmResourceId': element["vmResourceId"],
                    'vmRegionId': element["vmRegionId"],
                    'vmUser': element["vmUser"],
                    'instanceTypeId': element["instanceTypeId"],
                    'keyPairId': element["keyPairId"],
                    'openedPort': element["openedPort"],
            }
            if 'dockerPrivateRegistryIp' in element:
                   context['dockerPrivateRegistryIp'] = element["dockerPrivateRegistryIp"]
            if 'dockerPrivateRegistryPort' in element:
                    context['dockerPrivateRegistryPort'] =  element["dockerPrivateRegistryPort"]

            with open(fname, 'w') as f:
                    tf = render_template('vm-tpl.tf', context)
                    f.write(tf)




def get_docker_host_node_name(container_name, containers):
    for container in containers:
        if container["containerName"] == container_name:
                return container["dockerHostNodeName"]

    return ""

def create_services_tf(data_loaded):

    appName = data_loaded["name"]
    consulJoinIp = data_loaded["consulJoinIp"]

    infrastructure_dir = appName+"/infrastructure"
    services_dir = appName+"/services"
    if not os.path.exists(appName):
        os.makedirs(appName)
    if not os.path.exists(infrastructure_dir):
        os.makedirs(infrastructure_dir)
    if not os.path.exists(services_dir):
        os.makedirs(services_dir)

    networks_to_create = set()
    docker_providers = set()
    vmUser = {}
    for virtualMachine in data_loaded["virtualMachines"]:
        print(virtualMachine)
        dockerHostNodeName = virtualMachine["dockerHostNodeName"]
        vmUser[dockerHostNodeName] = virtualMachine["vmUser"] 
        
    for element in data_loaded["containers"]:
        print(element)
        fname = services_dir+"/"+appName+"-container-"+element["containerName"]+".tf"
        dockerHostNodeName = element["dockerHostNodeName"]
        context = {
                'appName': appName,
                'consulJoinIp': consulJoinIp,
                'containerName': element["containerName"],
                'imageName': element["imageName"],
                #'docker_host_ip': element["docker-host-ip"],
                'dockerHostNodeName': element["dockerHostNodeName"],
                'hostname': element["hostname"],
                'vmUser': vmUser[element["dockerHostNodeName"]],
        }
        docker_providers.add(element["dockerHostNodeName"])
        if 'imageTag' in element:
               context['imageTag'] = element["imageTag"]
        else:
               context['imageTag'] = "latest"
        if 'dockerPrivateRegistryIp' in element:
               context['dockerPrivateRegistryIp'] = element["dockerPrivateRegistryIp"]
        if 'dockerPrivateRegistryPort' in element:
               context['dockerPrivateRegistryPort'] =  element["dockerPrivateRegistryPort"]
        if 'dockerPrivateRegistryUser' in element:
               context['dockerPrivateRegistryUser'] = element["dockerPrivateRegistryUser"]
        if 'dockerPrivateRegistryPassword' in element:
               context['dockerPrivateRegistryPassword'] = element["dockerPrivateRegistryPassword"]
        if 'command' in element:
               context['command'] = json.dumps(element["command"])
        if 'entrypoint' in element:
               context['entrypoint'] = json.dumps(element["entrypoint"])
        if 'networks' in element:
               networks_dependencies = set()
               context['networks'] = element["networks"]
               for network in element["networks"]:
                        networks_dependencies.add("docker_network."+dockerHostNodeName+"-"+network)
                        if dockerHostNodeName+"-"+network not in networks_to_create:
                                networks_to_create.add(dockerHostNodeName+":"+network)
               context['networksDependencies'] = networks_dependencies
        if 'restart' in element:
               context['restart'] = element["restart"]
        if 'environment' in element:
               context['environment'] = json.dumps(element["environment"])
        if 'portMapping' in element:
               context['portMapping'] = element["portMapping"]
        if 'volumeMapping' in element:
               context['volumeMapping'] = element["volumeMapping"]
        if 'hostMapping' in element:
               for hostMappingElement in element["hostMapping"]:
                        if hostMappingElement['referencedContainerName']:
                                hostMappingElement['dockerHostNodeName'] = get_docker_host_node_name(hostMappingElement['referencedContainerName'], data_loaded["containers"])
               context['hostMapping'] = element["hostMapping"]
        if 'endpoints' in element:
               context['endpoints'] = element["endpoints"]
        if 'capabilities' in element:
               context['capabilities'] = element["capabilities"]
        if 'addConsulService' in element:
               context['addConsulService'] = element["addConsulService"]
        if 'consulServicePort' in element:
               context['consulServicePort'] = element["consulServicePort"]
        if 'addConsulTraefikRules' in element:
               context['addConsulTraefikRules'] = element["addConsulTraefikRules"]
        with open(fname, 'w') as f:
                tf = render_template('container-tpl.tf', context)
                f.write(tf)

    #Create common stuff section:
    context['dockerProviders'] = docker_providers
    
    fname = services_dir+"/"+appName+"-services-common.tf"
    with open(fname, 'w') as f:
        tf = render_template('services-common-tpl.tf', context)
        f.write(tf)

    #Create networks section:
    for network_to_create in networks_to_create:
        i = network_to_create.find(':')
        node_name =  network_to_create[:i]
        network_name = network_to_create[i+1:]
        fname = services_dir+"/"+appName+"-network-"+node_name+"-"+network_name+".tf"
        context = {
                'dockerHostNodeName': node_name,
                'networkName': network_name
        }

        with open(fname, 'w') as f:
                tf = render_template('network-tpl.tf', context)
                f.write(tf)


def create_adapt_tf(data_loaded):

    appName = data_loaded["name"]
    adapt_dir = appName+"-adapt"
    infrastructure_dir = adapt_dir+"/infrastructure"
    services_dir = adapt_dir+"/services"
    if not os.path.exists(appName):
        os.makedirs(appName)
    if not os.path.exists(adapt_dir):
        os.makedirs(adapt_dir)
    if not os.path.exists(infrastructure_dir):
        os.makedirs(infrastructure_dir)
    if not os.path.exists(services_dir):
        os.makedirs(services_dir)

    #Create folder for ssh key to connect to broker resources via keypairs
    keypairs_dir = "/home/ubuntu/terraform/keypairs/"+appName
    if not os.path.exists(keypairs_dir):
        os.makedirs(keypairs_dir)

    cloudbrokerEndpoint = data_loaded["cloudbrokerEndpoint"]
    cloudbrokerUsername = data_loaded["cloudbrokerUsername"]
    cloudbrokerPassword = data_loaded["cloudbrokerPassword"]

    vmUser = {}
    for virtualMachine in data_loaded["virtualMachines"]:
        print(virtualMachine)
        dockerHostNodeName = virtualMachine["dockerHostNodeName"]
        vmUser[dockerHostNodeName] = virtualMachine["vmUser"] 

    for element in data_loaded["virtualMachines"]:
        fname = infrastructure_dir+"/"+appName+"-vm-"+element["dockerHostNodeName"]+".tf"
        print(element)
        context = {
                'appName': appName,
                'dockerHostNodeName': element["dockerHostNodeName"],
                'cloudbrokerEndpoint': cloudbrokerEndpoint,
                'cloudbrokerUsername': cloudbrokerUsername,
                'cloudbrokerPassword': cloudbrokerPassword,
                'vmSoftwareId': element["vmSoftwareId"],
                'vmResourceId': element["vmResourceId"],
                'vmRegionId': element["vmRegionId"],
                'vmUser': element["vmUser"],
                'instanceTypeId': element["instanceTypeId"],
                'keyPairId': element["keyPairId"],
                'openedPort': element["openedPort"],
        }
        if 'dockerPrivateRegistryIp' in element:
               context['dockerPrivateRegistryIp'] = element["dockerPrivateRegistryIp"]
        if 'dockerPrivateRegistryPort' in element:
                context['dockerPrivateRegistryPort'] =  element["dockerPrivateRegistryPort"]
        with open(fname, 'w') as f:
            tf = render_template('adapt-vm-tpl.tf', context)
            f.write(tf)


    docker_providers = set()
    for element in data_loaded["containers"]:
        fname = services_dir+"/"+appName+"-container-"+element["dockerHostNodeName"]+".tf"
        print(element)
        dockerHostNodeName = element["dockerHostNodeName"]
        context['containerName'] = element["containerName"]
        context['imageName'] = element["imageName"]
        context['dockerHostNodeName'] = element["dockerHostNodeName"]
        context['hostname'] = element["hostname"]
        context['vmUser'] = vmUser[element["vmUser"]],
        docker_providers.add(element["dockerHostNodeName"])
        if 'imageTag' in element:
               context['imageTag'] = element["imageTag"]
        else:
               context['imageTag'] = "latest"
        if 'dockerPrivateRegistryUser' in element:
               context['dockerPrivateRegistryUser'] = element["dockerPrivateRegistryUser"]
        if 'dockerPrivateRegistryPassword' in element:
               context['dockerPrivateRegistryPassword'] = element["dockerPrivateRegistryPassword"]
        if 'command' in element:
               context['command'] = json.dumps(element["command"])
        if 'entrypoint' in element:
               context['entrypoint'] = json.dumps(element["entrypoint"])
        if 'restart' in element:
               context['restart'] = element["restart"]
        if 'environment' in element:
               context['environment'] = json.dumps(element["environment"])
        if 'portMapping' in element:
               context['portMapping'] = element["portMapping"]
        if 'hostMapping' in element:
               for hostMappingElement in element["hostMapping"]:
                        if hostMappingElement['referencedContainerName']:
                                hostMappingElement['dockerHostNodeName'] = get_docker_host_node_name(hostMappingElement['referencedContainerName'], data_loaded["containers"])
               context['hostMapping'] = element["hostMapping"]
        if 'endpoints' in element:
               context['endpoints'] = element["endpoints"]
        if 'capabilities' in element:
               context['capabilities'] = element["capabilities"]

        with open(fname, 'w') as f:
            tf = render_template('adapt-container-tpl.tf', context)
            f.write(tf)


def startMonitoring(body):

    repo_user = body["repository_user"]
    repo_pwd = body["repository_pwd"]
    repo_url = body["repository_url"]
    monitoring_host = body["monitoring_host"]
    monitoring_port = body["monitoring_port"]
    if not monitoring_host:
        print("*** MISSING MONITORING URL: simulating start of monitoring on a local dummy endpoint")
        monitoring_host = "localhost"
        monitoring_port = 8473
    elif not monitoring_port:
        print("*** MISSING MONITORING port: assuming port 80")
        monitoring_port = 80
    reqBody = {'user': repo_user, 'password': repo_pwd, 'appdescuri': repo_url, 'monid': 'test', 'status': 'test' }
    headers = {"Content-type": "application/json", "Accept": "*/*"}
    print("About to send POST request for monitoring. Url: " + monitoring_host+":"+monitoring_port)
    conn = http.client.HTTPConnection(monitoring_host, monitoring_port)
    conn.request("POST", "/monitoringmanager/api/applications", json.dumps(reqBody), headers)
    response = conn.getresponse()
    print(response.status, response.reason)
    data = response.read()
    conn.close()

def stopMonitoring(body, applicationInstanceId):

    repo_url = body["repository_url"]
    monitoring_host = body["monitoring_host"]
    monitoring_port = body["monitoring_port"]
    if not monitoring_host:
        print("*** MISSING MONITORING URL: simulating stop of monitoring on a local dummy endpoint")
        monitoring_host = "localhost"
        monitoring_port = 8473
    elif not monitoring_port:
        print("*** MISSING MONITORING port: assuming port 80")
        monitoring_port = 80

    reqBody = {'appdescuri': repo_url}
    headers = {"Content-type": "application/json", "Accept": "*/*"}
    print("About to send DELETE request for monitoring with applicationInstanceId " + applicationInstanceId)
    conn = http.client.HTTPConnection(monitoring_host, monitoring_port)
    requestPath = "/monitoringmanager/api/applications/"+applicationInstanceId
    conn.request("DELETE", requestPath, json.dumps(reqBody), headers)
    response = conn.getresponse()
    print(response.status, response.reason)
    data = response.read()
    conn.close()


########################





if __name__ == "__main__":
    # Only for debugging while developing
    app.run(host='0.0.0.0', debug=True, port=8089)

